///////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2020 Martin Wölzer
//
// This code is licensed under the MIT License (MIT).
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
///////////////////////////////////////////////////////////////////////////////

#ifndef CLAPP_OPTION_HPP
#define CLAPP_OPTION_HPP

#include <clapp/option.h>
#include <clapp/type_traits.h>
#include <clapp/value.h>

#include <numeric>
#include <sstream>

namespace clapp::option {
template <typename T>
basic_param_option_t<T>::~basic_param_option_t() = default;

template <typename T>
basic_vector_param_option_t<T>::~basic_vector_param_option_t() = default;

template <typename T, T DefaultValue>
basic_option_t<T, DefaultValue>::~basic_option_t() = default;

template <int ExitCode>
basic_help_option_t<ExitCode>::~basic_help_option_t() = default;
}  // namespace clapp::option

template <typename T>
inline void clapp::option::gen_opt_conf_process_params(
    [[maybe_unused]] opt_params_t<T>& opt_params) {}

template <typename T, typename Param>
void clapp::option::gen_opt_conf_process_params(opt_params_t<T>& opt_params,
                                                Param&& param) {
    if constexpr (type_traits::has_append_restriction_text<
                      std::decay_t<Param>>::value) {
        opt_params.restrictions.push_back(
            std::string{param.append_restriction_text()});
    }
    if constexpr (type_traits::has_default_value<std::decay_t<Param>>::value) {
        opt_params.default_value = param.default_value();
    }
    if constexpr (type_traits::has_validate<std::decay_t<Param>, T>::value) {
        opt_params.validate_funcs.push_back(
            [param_capt = std::forward<Param>(param)](
                const T& value, const std::string& option_string) {
                param_capt.validate(value, option_string);
            });
    }
    if constexpr (type_traits::has_found<std::decay_t<Param>>::value) {
        opt_params.found.push_back(std::forward<Param>(param));
    }
    if constexpr (std::is_same_v<std::decay_t<Param>,
                                 parser::types::purpose_t>) {
        opt_params.purpose = std::forward<Param>(param);
    }
}

template <typename T, typename Param, typename... Params>
void clapp::option::gen_opt_conf_process_params(opt_params_t<T>& opt_params,
                                                Param&& param,
                                                Params&&... parameters) {
    gen_opt_conf_process_params(opt_params, std::forward<Param>(param));
    gen_opt_conf_process_params(opt_params,
                                std::forward<Params>(parameters)...);
}

template <typename ShortOptionFuncT>
std::vector<clapp::parser::types::basic_short_opt_conf_t<ShortOptionFuncT>>
clapp::option::gen_short_option(ShortOptionFuncT&& sof,
                                const std::vector<char>& short_option) {
    std::vector<parser::types::basic_short_opt_conf_t<ShortOptionFuncT>> ret;
    for (const auto opt : short_option) {
        check_short_option(opt);
        ret.emplace_back(
            parser::types::basic_short_opt_conf_t<ShortOptionFuncT>{opt, sof});
    }
    return ret;
}

template <typename LongOptionFuncT>
std::vector<clapp::parser::types::basic_long_opt_conf_t<LongOptionFuncT>>
clapp::option::gen_long_option(LongOptionFuncT&& lof,
                               const std::vector<std::string>& long_option) {
    std::vector<parser::types::basic_long_opt_conf_t<LongOptionFuncT>> ret;
    for (const auto& opt : long_option) {
        check_long_option(opt);
        ret.emplace_back(
            parser::types::basic_long_opt_conf_t<LongOptionFuncT>{opt, lof});
    }
    return ret;
}

template <typename T, typename ValueFunc>
clapp::parser::types::validate_value_func_t
clapp::option::gen_opt_validate_value_func(
    ValueFunc&& value_func, has_value_func_t&& hvf,
    std::vector<typename opt_params_t<T>::validate_func_t>&& validate_funcs) {
    if (!validate_funcs.empty()) {
        return [value_func_capt = std::forward<ValueFunc>(value_func),
                has_value_func = std::move(hvf),
                validate_funcs_capt = std::move(validate_funcs)](
                   const std::string& option_string) {
            if (has_value_func()) {
                if constexpr (std::is_same_v<ValueFunc,
                                             vector_value_func_t<T>>) {
                    for (const auto& value : value_func_capt()) {
                        for (const auto& func : validate_funcs_capt) {
                            func(value, option_string);
                        }
                    }
                } else if constexpr (std::is_same_v<ValueFunc,
                                                    value_func_t<T>>) {
                    const T value{value_func_capt()};
                    for (const auto& func : validate_funcs_capt) {
                        func(value, option_string);
                    }
                }
            }
        };
    }
    return [](const std::string& /*option_string*/) {};
}

template <typename T>
inline std::vector<std::string> clapp::option::gen_string_vec(
    const std::vector<T>& vec) {
    return std::vector<std::string>{vec.begin(), vec.end()};
}

template <>
inline std::vector<std::string> clapp::option::gen_string_vec(
    const std::vector<std::string>& vec) {
    return vec;
}

template <typename T, typename OptConf, typename Callbacks, typename T1,
          typename... Params>
clapp::option::opt_conf_container_t<T, OptConf> clapp::option::gen_opt_conf(
    Callbacks&& callbacks, T1&& single_option, const std::string& description,
    Params&&... parameters) {
    if constexpr (clapp::type_traits::is_vector<std::decay_t<T1>>::value) {
        if constexpr (std::is_same_v<std::decay_t<T1>, std::vector<char>>) {
            return gen_opt_conf1<T, OptConf>(
                std::forward<Callbacks>(callbacks), std::vector<std::string>{},
                std::forward<T1>(single_option), description,
                std::forward<Params>(parameters)...);
        } else {
            return gen_opt_conf1<T, OptConf>(
                std::forward<Callbacks>(callbacks),
                gen_string_vec(std::forward<T1>(single_option)),
                std::vector<char>{}, description,
                std::forward<Params>(parameters)...);
        }
    } else if constexpr (std::is_same_v<std::decay_t<T1>, char>) {
        return gen_opt_conf1<T, OptConf>(
            std::forward<Callbacks>(callbacks), std::vector<std::string>{},
            std::vector<char>{std::forward<T1>(single_option)}, description,
            std::forward<Params>(parameters)...);
    } else {
        return gen_opt_conf1<T, OptConf>(
            std::forward<Callbacks>(callbacks),
            std::vector<std::string>{{std::forward<T1>(single_option)}},
            std::vector<char>{}, description,
            std::forward<Params>(parameters)...);
    }
}

template <typename T, typename OptConf, typename Callbacks, typename T1,
          typename T2, typename... Params>
clapp::option::opt_conf_container_t<T, OptConf> clapp::option::gen_opt_conf(
    Callbacks&& callbacks, T1&& long_option, T2&& short_option,
    const std::string& description, Params&&... parameters) {
    if constexpr (clapp::type_traits::is_vector<std::decay_t<T1>>::value &&
                  clapp::type_traits::is_vector<std::decay_t<T2>>::value) {
        return gen_opt_conf1<T, OptConf>(
            std::forward<Callbacks>(callbacks),
            gen_string_vec(std::forward<T1>(long_option)),
            std::forward<T2>(short_option), description,
            std::forward<Params>(parameters)...);
    }
    if constexpr (clapp::type_traits::is_vector<std::decay_t<T1>>::value) {
        return gen_opt_conf1<T, OptConf>(
            std::forward<Callbacks>(callbacks),
            gen_string_vec(std::forward<T1>(long_option)),
            std::vector<char>{std::forward<T2>(short_option)}, description,
            std::forward<Params>(parameters)...);
    }
    if constexpr (clapp::type_traits::is_vector<std::decay_t<T2>>::value) {
        return gen_opt_conf1<T, OptConf>(
            std::forward<Callbacks>(callbacks),
            std::vector<std::string>{std::forward<T1>(long_option)},
            std::forward<T2>(short_option), description,
            std::forward<Params>(parameters)...);
    }
    return gen_opt_conf1<T, OptConf>(
        std::forward<Callbacks>(callbacks),
        std::vector<std::string>{std::forward<T1>(long_option)},
        std::vector<char>{std::forward<T2>(short_option)}, description,
        std::forward<Params>(parameters)...);
}

template <typename T, typename OptConf, typename Callbacks, typename... Params>
clapp::option::opt_conf_container_t<T, OptConf> clapp::option::gen_opt_conf1(
    Callbacks&& callbacks, const std::vector<std::string>& long_option,
    const std::vector<char>& short_option, const std::string& description,
    Params&&... parameters) {
    opt_params_t<T> opt_params;
    if (std::is_same_v<OptConf, parser::types::opt_vector_param_conf_t>) {
        opt_params.restrictions.push_back("vector option");
    }

    gen_opt_conf_process_params(opt_params,
                                std::forward<Params>(parameters)...);

    const std::string purpose{
        parser::types::to_string_view(opt_params.purpose).value()};

    std::string restriction{std::accumulate(
        opt_params.restrictions.begin(), opt_params.restrictions.end(), purpose,
        [](const std::string& lhs, const std::string& rhs) -> std::string {
            const std::string sep{lhs.length() > 0 && rhs.length() > 0 ? ", "
                                                                       : ""};
            return lhs + sep + rhs;
        })};
    if (!restriction.empty()) {
        restriction = " (" + restriction + ")";
    }

    return opt_conf_container_t<T, OptConf>{
        gen_opt_conf2<T, OptConf>(
            std::forward<Callbacks>(callbacks), long_option, short_option,
            std::move(opt_params.validate_funcs), description + restriction,
            opt_params.purpose),
        opt_params.default_value, std::move(opt_params.found)};
}

template <typename T, typename OptConf, typename Callbacks>
OptConf clapp::option::gen_opt_conf2(
    Callbacks&& callbacks, const std::vector<std::string>& long_option,
    const std::vector<char>& short_option,
    std::vector<typename opt_params_t<T>::validate_func_t>&& validate_funcs,
    const std::string& description, parser::types::purpose_t purpose) {
    const std::string option_string{
        OptConf::create_option_string(short_option, long_option)};

    using short_opt_func_t = typename std::decay_t<Callbacks>::short_opt_func_t;
    using short_opt_conf_vec_t =
        std::vector<parser::types::basic_short_opt_conf_t<short_opt_func_t>>;
    short_opt_conf_vec_t short_options{
        gen_short_option(std::move(callbacks.soh), short_option)};

    using long_opt_func_t = typename std::decay_t<Callbacks>::long_opt_func_t;
    using long_opt_conf_vec_t =
        std::vector<parser::types::basic_long_opt_conf_t<long_opt_func_t>>;
    long_opt_conf_vec_t long_options{
        gen_long_option(std::move(callbacks.loh), long_option)};

    using validate_value_func_t = parser::types::validate_value_func_t;
    validate_value_func_t validate_value_func{gen_opt_validate_value_func<T>(
        std::move(callbacks.value), std::move(callbacks.has_value),
        std::move(validate_funcs))};

    return OptConf{std::move(short_options),
                   std::move(long_options),
                   std::move(callbacks.given),
                   std::move(validate_value_func),
                   option_string,
                   description,
                   purpose};
}

template <typename T>
template <typename... Params>
clapp::option::basic_param_option_t<T>::basic_param_option_t(
    parser::basic_option_container_t& container, Params... parameters) {
    opt_conf_container_t<T, opt_conf_t> conf{gen_opt_conf<T, opt_conf_t>(
        create_callbacks(this), std::forward<Params>(parameters)...)};
    container.reg(std::move(conf.opt_conf));
    _value = conf.default_value;
    _found = std::move(conf.found);
}

template <typename T>
typename clapp::option::basic_param_option_t<T>::callbacks_t
clapp::option::basic_param_option_t<T>::create_callbacks(
    basic_param_option_t<T>* inst) {
    return callbacks_t{
        [inst](const std::string_view option, const std::string_view param) {
            return inst->found_entry("--" + std::string{option}, param);
        },
        [inst](const char option, const std::string_view param) {
            return inst->found_entry("-" + std::string{option}, param);
        },
        [inst]() { return inst->given(); },
        [inst]() { return static_cast<bool>(*inst); },
        [inst]() { return inst->value(); }};
}

template <typename T>
clapp::value::found_func_t::ret_t
clapp::option::basic_param_option_t<T>::found_entry(
    const std::string& option, const std::string_view param) {
    for (auto& found_func : _found) {
        const clapp::value::found_func_t::ret_t ret{found_func.found(option)};
        if (ret) {
            return ret;
        }
    }
    _given = true;
    _value = value::convert_value<T>(param);
    return {};
}

template <typename T>
constexpr clapp::option::basic_param_option_t<T>::operator bool()
    const noexcept {
    return _value.has_value();
}

template <typename T>
constexpr bool clapp::option::basic_param_option_t<T>::has_value()
    const noexcept {
    return _value.has_value();
}

template <typename T>
T clapp::option::basic_param_option_t<T>::value() const {
    if (!_value) {
        throw clapp::value_undefined_t{"Requested value is not defined."};
    }
    return T{_value.value()};
}

template <typename T>
constexpr bool clapp::option::basic_param_option_t<T>::given() const noexcept {
    return _given;
}

template <typename T>
template <typename... Params>
clapp::option::basic_vector_param_option_t<T>::basic_vector_param_option_t(
    parser::basic_option_container_t& container, Params... parameters) {
    opt_conf_container_t<T, opt_conf_t> conf{gen_opt_conf<T, opt_conf_t>(
        create_callbacks(this), std::forward<Params>(parameters)...)};
    if (conf.default_value) {
        std::stringstream string_stream;
        string_stream << "No default value for vector based param option '"
                      << conf.opt_conf.create_basic_option_string()
                      << "' possible.";
        throw clapp::option_param_exception_t{string_stream.str()};
    }
    container.reg(std::move(conf.opt_conf));
    _found = std::move(conf.found);
}

template <typename T>
typename clapp::option::basic_vector_param_option_t<T>::callbacks_t
clapp::option::basic_vector_param_option_t<T>::create_callbacks(
    basic_vector_param_option_t<T>* inst) {
    return callbacks_t{
        [inst](const std::string_view option, const std::string_view param) {
            return inst->found_entry("--" + std::string{option}, param);
        },
        [inst](const char option, const std::string_view param) {
            return inst->found_entry("-" + std::string{option}, param);
        },
        [inst]() { return inst->given(); },
        [inst]() { return static_cast<bool>(*inst); },
        [inst]() { return inst->value(); }};
}

template <typename T>
clapp::value::found_func_t::ret_t
clapp::option::basic_vector_param_option_t<T>::found_entry(
    const std::string& option, const std::string_view param) {
    for (auto& found_func : _found) {
        const clapp::value::found_func_t::ret_t ret{found_func.found(option)};
        if (ret) {
            return ret;
        }
    }
    _given = true;
    _value.push_back(value::convert_value<T>(param));
    return {};
}

template <typename T>
inline clapp::option::basic_vector_param_option_t<T>::operator bool()
    const noexcept {
    return !_value.empty();
}

template <typename T>
inline bool clapp::option::basic_vector_param_option_t<T>::has_value()
    const noexcept {
    return !_value.empty();
}

template <typename T>
std::vector<T> clapp::option::basic_vector_param_option_t<T>::value() const {
    return _value;
}

template <typename T>
constexpr bool clapp::option::basic_vector_param_option_t<T>::given()
    const noexcept {
    return _given;
}

template <typename T, T DefaultValue>
template <typename... Params>
clapp::option::basic_option_t<T, DefaultValue>::basic_option_t(
    parser::basic_option_container_t& container, callbacks_t&& callbacks,
    Params&&... parameters) {
    opt_conf_container_t<T, opt_conf_t> conf{gen_opt_conf<T, opt_conf_t>(
        std::move(callbacks), std::forward<Params>(parameters)...)};
    container.reg(std::move(conf.opt_conf));
    if (conf.default_value) {
        _value = conf.default_value.value();
    }
    _found = std::move(conf.found);
}

template <typename T, T DefaultValue>
T clapp::option::basic_option_t<T, DefaultValue>::value() const {
    return _value;
}

template <typename T, T DefaultValue>
bool clapp::option::basic_option_t<T, DefaultValue>::given() const {
    return _given;
}

template <typename T, T DefaultValue>
constexpr clapp::option::basic_option_t<T, DefaultValue>::operator bool()
    const noexcept {
    return _value != DefaultValue;
}

template <typename T, T DefaultValue>
constexpr bool clapp::option::basic_option_t<T, DefaultValue>::has_value()
    const noexcept {
    return true;
}

template <typename... Params>
clapp::option::bool_option_t::bool_option_t(
    parser::basic_option_container_t& container, Params... parameters)
    : clapp::option::basic_option_t<bool, false>{
          container, create_callbacks(this),
          std::forward<Params>(parameters)...} {
    if (!_value) {
        _value = false;
    }
}

template <int ExitCode>
template <typename... Params>
clapp::option::basic_help_option_t<ExitCode>::basic_help_option_t(
    parser::basic_option_container_t& container, Params... parameters)
    : bool_option_t{container, std::forward<Params>(parameters)...,
                    gen_func_print_help_and_req_exit(container)} {}

template <int ExitCode>
clapp::value::found_func_t
clapp::option::basic_help_option_t<ExitCode>::gen_func_print_help_and_req_exit(
    parser::basic_option_container_t& container) {
    return container.get_parser().gen_func_print_help_and_req_exit(ExitCode);
}

template <typename... Params>
clapp::option::count_option_t::count_option_t(
    parser::basic_option_container_t& container, Params... parameters)
    : clapp::option::basic_option_t<std::uint32_t, 0U>{
          container, create_callbacks(this),
          std::forward<Params>(parameters)...} {
    if (!_value) {
        _value = 0;
    }
}

#endif
